
// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize) ;
}

Format BlockRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockRangeRep :: ker_glb(const int glb, int* sub) const {

  *sub = glb % blockSize ;
  return glb / blockSize ;
}

void BlockRangeRep :: location(Location* loc, const int glb) const {

  loc->subs[0] = glb % blockSize ;
  loc->crd     = glb / blockSize ;
  loc->dim     = dimension ;
}

RangeRep* BlockRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

int BlockRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  const int blockStart  = ker_glb * blockSize ;

  *sub = 0 ;
  if(ker_glb <= ker_hi) {
    if(blockStart + blockSize > extent)
      *count = extent - blockStart ;
    else
      *count = blockSize ;
  }
  else
    *count = 0 ;

  return blockStart ;
}

int BlockRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = 1 ;
  return 1 ;
}

RangeRep* BlockRangeRep :: subker(int* ker_bas, int* ker_str) const {

  *ker_bas = 0 ;
  *ker_str = 1 ;

  return kernel->subrng(ker_hi + 1, 0, 1) ;
}

int BlockRangeRep :: width() const {return blockSize ; }

int BlockRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb * blockSize + sub ;
}

void BlockRangeRep :: show() {
  AdlibPrintf("BlockRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, blockSize) ;
}

RangeRep* BlockSubRangeRep :: tem() {return parent ; }
int BlockSubRangeRep :: str() const {return stride ; }
int BlockSubRangeRep :: bas() const {return base ; }

Format BlockSubRangeRep :: format() const {
  return DIST_BLOCK ;
}

int BlockSubRangeRep :: ker_glb(const int glb, int* sub) const {
  const int tem = base + stride * glb ;

  *sub = tem % blockSize ;
  return tem / blockSize ;
}

void BlockSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  loc->subs[0] = tem % blockSize ;
  loc->crd     = tem / blockSize ;
  loc->dim     = dimension ;
}

RangeRep* BlockSubRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

int BlockSubRangeRep :: block(const int ker_glb, int* sub, int* count) const {

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    const int blockStart = ker_glb * blockSize ;

    const int blockStartOffset  = blockStart - base ;
    const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

    int i_hi, i_lo ;
    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }
   
    *sub   = base + stride * i_lo - blockStart ;
    *count = i_hi - i_lo + 1 ;

    return i_lo ;
  }
  else {
    //*sub   = base ;  // What is the logic of this?
    *sub   = 0 ;
    *count = 0 ;

    return 0 ;
  }
}

int BlockSubRangeRep :: block_step(int* sub_stp) const {

  *sub_stp = stride ;
  return 1 ;
}

RangeRep* BlockSubRangeRep :: subker(int* ker_bas, int* ker_str) const {

  if(stride > 0) {
    *ker_bas = ker_lo ;
    *ker_str = 1 ;
  }
  else {
    *ker_bas = ker_hi ;
    *ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, *ker_bas, *ker_str) ;
}

int BlockSubRangeRep :: width() const {return blockSize ; }

int BlockSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (ker_glb * blockSize + sub - base) / stride ;
}

void BlockSubRangeRep :: show() {
  AdlibPrintf("BlockSubRangeRep :: show: "
              "size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range of array with ghost areas.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedBlockRangeRep :: subrng(const int _extent,
                                          const int _base, const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent, _base, _stride, this,
                                      kernel, blockSize, wlo, whi) ;
}

RangeRep* ExtendedBlockRangeRep :: shell() const {
  return new ExtendedCollapsedRangeRep(blockSize, wlo, whi) ;
}

int ExtendedBlockRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedBlockRangeRep :: offset(const Location* i) const {
  return i->subs [0] + wlo ;
}

int ExtendedBlockRangeRep :: width() const {
  return wlo + blockSize + whi ;
}

RangeRep* ExtendedBlockSubRangeRep :: subrng(const int _extent,
                                             const int _base,
                                             const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent,
				      base + stride * _base, stride * _stride,
				      parent, kernel, blockSize, wlo, whi) ;
}

RangeRep* ExtendedBlockSubRangeRep :: shell() const {
  return new ExtendedCollapsedRangeRep(blockSize, wlo, whi) ;
}

int ExtendedBlockSubRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedBlockSubRangeRep :: offset(const Location* i) const {
  return i->subs [0] + wlo ;
}

int ExtendedBlockSubRangeRep :: width() const {
  return wlo + blockSize + whi ;
}


MEMORY_POOL_DEFN(ExtendedBlockRangeRep)

MEMORY_POOL_DEFN(ExtendedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockSubRangeRep':                                   //
//      Block-distributed ranges of array with ``packed'' layout.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent,
				    base + stride * _base, stride * _stride,
				    parent, kernel, blockSize, packing) ;
}

RangeRep* PackedBlockSubRangeRep :: shell() const {
  AdlibError("PackedBlockSubRangeRep :: shell: "
             "member undefined on this range") ;
  return 0 ;
}

int PackedBlockSubRangeRep :: disp(const int sub) const {
  return sub / packing ;
}

int PackedBlockSubRangeRep :: step(const int sub_stp) const {
  return sub_stp / packing ;
}

int PackedBlockSubRangeRep :: offset(const Location* i) const {
  return i->subs [0] / packing ;
}

int PackedBlockSubRangeRep :: width() const {
  return (blockSize + packing - 1) / packing ;
}

MEMORY_POOL_DEFN(PackedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `BlockCyclicRangeRep', `BlockCyclicSubRangeRep':            //
//     Block-cyclically distributed ranges.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockCyclicRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent, _base, _stride, this,
                                    kernel, blockSize) ;
}

void BlockCyclicRangeRep :: location(Location* loc, const int glb) const {
  const int numProcs = dimension.size() ;

  const int ker = glb / blockSize ;

  loc->subs[0] = glb % blockSize ;
  loc->subs[1] = ker / numProcs ;
  loc->crd     = ker % numProcs ;
  loc->dim     = dimension ;
}

int BlockCyclicRangeRep :: offset(const Location* i) const {
  return i->subs [0] + blockSize * i->subs [1] ;
}

int BlockCyclicRangeRep :: volume() const {
  return blockSize * kernel->width() ;
}

RangeRep* BlockCyclicSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent,
			            base + stride * _base, stride * _stride,
			            parent, kernel, blockSize) ;
}

void BlockCyclicSubRangeRep :: location(Location* loc, const int glb) const {
  const int numProcs = dimension.size() ;

  const int tem = base + stride * glb ;
  const int ker = tem / blockSize ;

  loc->subs[0] = tem % blockSize ;
  loc->subs[1] = ker / numProcs ;
  loc->crd     = ker % numProcs ;
  loc->dim     = dimension ;
}

int BlockCyclicSubRangeRep :: offset(const Location* i) const {
  return i->subs [0] + blockSize * i->subs [1] ;
}

int BlockCyclicSubRangeRep :: volume() const {
  return blockSize * kernel->width() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockCyclicSubRangeRep':                               //
//      Block-cyclically distributed range of array with ``packed''    //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockCyclicSubRangeRep :: subrng(const int _extent,
                                                 const int _base,
                                                 const int _stride) {
  return new PackedBlockCyclicSubRangeRep(_extent,
			                  base + stride * _base,
                                          stride * _stride,
			                  parent, kernel, blockSize, packing) ;
}

void PackedBlockCyclicSubRangeRep :: location(Location* loc,
                                              const int glb) const {
  const int numProcs = dimension.size() ;

  const int tem = base + stride * glb ;
  const int ker = tem / blockSize ;

  loc->subs[0] = tem % blockSize ;
  loc->subs[1] = ker / numProcs ;
  loc->crd     = ker % numProcs ;
  loc->dim     = dimension ;
}

int PackedBlockCyclicSubRangeRep :: offset(const Location* i) const {
  return i->subs [0] / packing + width() * i->subs [1] ;
}

int PackedBlockCyclicSubRangeRep :: volume() const {
  return width() * kernel->width() ;
}


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)
//
//   Simplified sub-range classes for contiguous subranges.
//
//   Extended block-cyclic subclasses.
//
//   `show' methods for subclasses.

